functor LatticeGraph (val dim : int; structure Ring : SEMIRING) =
    struct
       structure Mat = 
       Matrix(structure R = Ring; val rows = dim; val cols = dim)

       fun singleEdgeExtend (H,V,q) D =
           let
               val sum = Mat.sum
               val prod = Mat.prod
               val tran = Mat.tran
               val F = tran(prod(tran D,V))
               val E = let val (qi,qj) = q 
                       in 
                           Mat.tabulate 
                               (fn (i,j) => if i = qi andalso (j > qj) 
                                            then Mat.sub(H,(qj,j)) 
                                            else if j = qj andalso (i > qi)
                                            then Mat.sub(V,(qi,i))
                                            else Ring.zero)
                       end
           in 
               sum(E,sum(D,sum(F,prod(D,H))))
           end

       (* Recursively compute the distance to points in subsquare X with 
          NW = (i1,j1) SE = (i2,j2) by splitting X into four subsquares *)

       fun shortestPaths (H,V,q) =
           let 
               fun paths Dist (i1,j1) (i2,j2) =
                   if i2 = i1 then Dist
                   else 
                       if i2 = i1+1 then singleEdgeExtend (H,V,q) Dist
                       else
                           let 
                               val m = (i2+1-i1) div 2 
                               val A = paths Dist (i1,j1) (i1+m-1,j1+m-1) 
                               val B = paths A (i1+m,j1) (i2,j1+m-1)
                               val D = paths A (i1,j1+m) (i1+m-1,j2)
                               val C = paths D (i1+m,j1+m) (i2,j2)
                           in
                               C 
                           end 
               val (n,_) = Mat.dim
           in
               paths Mat.zero (0,0) (n-1,n-1)
           end
    end
