(* CM.make "rytter.cm"; *)
val _ = Control.Print.printDepth := 30;

structure G : CNF =
  struct
    datatype syncat = S | A | B | C | D | E | F
    datatype lexcat =     a | b | c | d | e
    datatype rule = SynRule of syncat * syncat * syncat 
                  | LexRule of syncat * lexcat
    val rules = [ SynRule(S,A,B), SynRule(A,C,D), SynRule(B,E,F),
                  SynRule(C,B,A), SynRule(C,A,E), SynRule(E,C,B),
                  SynRule(E,B,C), SynRule(F,B,E),
                  LexRule(A,a), LexRule(B,b), 
                  LexRule(C,c), LexRule(D,d), LexRule(E,e) 
                ]
    val syncats = [S, A, B, C, D, E, F]
    fun nth i = List.nth(syncats,i)
  end;

datatype lexcat = datatype G.lexcat;
structure Cyk = Rytter(val word = [b, c, d, d, c, b, b, e] : G.lexcat list;
                       structure G = G);
val M = Cyk.Valid(0,7); 
val ppM = Cyk.showPM M;
(* val ppM =
  #[#[[B],[E],[C],[A],[],[C],[E],[S]],
    #[[],[C],[A],[],[],[],[],[]],
    #[[],[],[D],[],[],[],[],[]],
    #[[],[],[],[D],[],[],[],[]],
    #[[],[],[],[],[C],[E],[],[B]],
    #[[],[],[],[],[],[B],[],[]],
    #[[],[],[],[],[],[],[B],[F]],
    #[[],[],[],[],[],[],[],[E]]]
*)
val (H,V,q) = Cyk.latticeGraph(M, (0,7))
val ppLG = Cyk.showLG (H,V,q);
(* val it =
 (#[#[[],[(B,E)],[(B,C)],[],[],[],[],[]],
    #[[],[],[(C,A)],[],[],[],[],[]],
    #[[],[],[],[(C,A)],[],[],[],[]],
    #[[],[],[],[],[(B,E)],[(A,C),(B,F)],[],[(A,S),(C,E)]],
    #[[],[],[],[],[],[(A,S),(C,E)],[],[]],
    #[[],[],[],[],[],[],[(A,S),(C,E)],[(E,B)]],
    #[[],[],[],[],[],[],[],[(A,C),(B,F)]],
    #[[],[],[],[],[],[],[],[]]
   ],
  #[#[[],[],[],[],[],[],[],[]],
    #[[(A,C),(C,E),(E,F)],[],[],[],[],[],[],[]],
    #[[(F,B)],[(B,E),(D,A)],[],[],[],[],[],[]],
    #[[(B,E),(D,A)],[(B,S),(E,C)],[],[],[],[],[],[]],
    #[[(B,S),(E,C)],[],[],[],[],[],[],[]],
    #[[],[],[],[],[(B,E),(D,A)],[],[],[]],
    #[[(B,E),(D,A)],[],[],[],[(F,B)],[(A,C),(C,E),(E,F)],[],[]],
    #[[(F,B)],[],[],[],[],[],[(A,C),(C,E),(E,F)],[]]
   ],
  (3,3))
*)
fun ppDist (q1,q2) = 
    let val n = 8
        val Vtr = Cyk.LG.Mat.tabulate 
                      (fn(i,j) => Cyk.LG.Mat.sub(V,(n-1-i,n-1-j)))  
        val qtr = (n-1-q1,q2)   
        val Dist = Cyk.LG.shortestPaths (H,Vtr,qtr) 
        fun swap X = Cyk.LG.Mat.tabulate 
                         (fn(i,j) => Cyk.LG.Mat.sub(X,(n-1-i,j)))  
    in 
        Cyk.showDist (swap Dist)
    end
(* val ppDist q =
 #[#[[],[],[],[(B,E),(D,A),(E,E)],[(B,E)],[(B,F),(D,C)],[(A,E),(D,E)],
                                                    [(C,E),(D,S),(E,F)]],
   #[[],[],[],[(B,S),(E,C)],[(B,C)],[(B,E)],[(A,C)],[(B,B),(C,C),(E,E)]],
   #[[],[],[],[],[],[],[],[]],
   #[[],[],[],[],[(B,E)],[(A,C),(B,F)],[(A,E)],[(A,S),(C,E)]],
   #[[],[],[],[],[],[],[],[]],
   #[[],[],[],[],[],[],[],[]],
   #[[],[],[],[],[],[],[],[]],
   #[[],[],[],[],[],[],[],[]]
  ]
*)
