signature CNF = 
    sig
        eqtype syncat
        eqtype lexcat
        datatype rule = SynRule of syncat * syncat * syncat 
                      | LexRule of syncat * lexcat
        val rules : rule list
        val syncats : syncat list
        val nth : int -> syncat
    end
functor Rytter(structure G:CNF; val word :G.lexcat list) = 
struct 
   val s = List.length G.syncats
   structure NTrel = Matrix(structure R = Boole; val rows = s; val cols = s)
   structure NTset = Matrix(structure R = Boole; val rows = 1; val cols = s)

   fun apply (R:NTrel.t, A:NTset.t) = NTset.prod(A,R)

   val n = List.length word
   structure PM = Matrix(structure R = NTset; val rows = n; val cols = n)
   structure LG = LatticeGraph(val dim = n; structure Ring = NTrel) 

   structure MatR = Matrices(NTrel:SEMIRING) 

   fun syncatToNTset (X:G.syncat) = 
       NTset.tabulate (fn (i,j) => if G.nth(j) = X then 1 else 0)

   fun lexcatToNTset (a:G.lexcat) = 
       let
           val cats = List.map (fn (G.LexRule(A,_)) => A)
                               (List.filter (fn (G.LexRule(_,b)) => a = b 
                                                 | _ => false) G.rules)
           fun add (X::Xs) = NTset.sum(syncatToNTset X, add Xs)
             | add [] = NTset.zero
       in
           add cats 
       end

   local 
       val makeNTset = NTset.tabulate
       val makeNTrel = NTrel.tabulate
       fun isSynRule (k,i,j) = 
           let 
               val rule = G.SynRule(G.nth k,G.nth i,G.nth j)
           in 
               List.exists (fn r => r = rule) G.rules
           end
   in
   val PH = Vector.tabulate 
                (s, fn C => makeNTrel (fn (B,A) => if isSynRule(A,B,C) 
                                                   then 1 else 0))
   val PV = Vector.tabulate 
                (s, fn B => makeNTrel (fn (C,A) => if isSynRule(A,B,C) 
                                                   then 1 else 0))
   fun Valid (n1,n2) =
       if n2-n1 >= 1 
       then 
           let
               val n3 = (n2+1 - n1) div 2 
               val sub = PM.sub
               val _ = TextIO.print ("Valid(" ^ (Int.toString n1) 
                                     ^ ", "   ^ (Int.toString n2) ^ ")\n")
           in 
               if 2 * n3 = (n2+1 - n1) 
               then
                   let
                       val NW = Valid (n1,n1+n3-1)
                       val SE = Valid (n1+n3,n2)
                       val M = PM.sum(NW,SE)
                       val (H,V,(q1,q2)) = latticeGraph(M,(n1,n2))
                   in
                       if n2-n1 = 1 
                       then
                           let 
                               fun valid(i,j) =
                                   if (i,j) = (n1,n1) orelse (i,j) = (n2,n2)
                                   then lexcatToNTset (List.nth(word,i)) 
                                   else if (i,j) = (n1,n2) 
                                   then NTset.sum(apply(LG.Mat.sub(H,(i,j)),
                                                        sub(M,(i,i))),
                                                  apply(LG.Mat.sub(V,(j,i)),
                                                        sub(M,(j,j))))
                                   else NTset.zero
                           in
                               PM.tabulate valid
                           end
                       else
                           let (* correct row numbers for vertical edges and source node *)
                               val Vtr = LG.Mat.tabulate 
                                             (fn(i,j) => LG.Mat.sub(V,(n-1-i,n-1-j)))  
                               val qtr = (n-1-q1,q2)   
                               val Dist = LG.shortestPaths (H,Vtr,qtr) 
                               fun valid(i,j) = 
                                   if (n1 <= Int.min(i,j) andalso Int.max(i,j) < n1+n3)
                                      orelse 
                                      (n1+n3 <= Int.min(i,j) andalso Int.max(i,j) <= n2)
                                   then sub(M,(i,j))
                                   else if n1 <= i andalso i < n1+n3 
                                           andalso n1+n3 <= j andalso j <= n2
                                   then apply(LG.Mat.sub(Dist,(n-1-i,j)),  
                                              PM.sub(M,(q1,q2))) 
                                   else NTset.zero
                           in
                               PM.tabulate valid
                           end
                   end
               else
                   raise Size
           end
       else
           let 
               val _ = TextIO.print ("Valid(" ^ (Int.toString n1) 
                                     ^ ", "   ^ (Int.toString n2) ^ ")\n")
               fun valid(i,j) = 
                   if (i,j) = (n1,n1) 
                   then lexcatToNTset (List.nth(word,i)) 
                   else NTset.zero
           in
               PM.tabulate valid
           end
   and latticeGraph (M:PM.t,(n1,n2):int*int) =
       let 
           val _ = TextIO.print ("latticeGraph(M,(" ^ (Int.toString n1) 
                                 ^ "," ^ (Int.toString n2) ^"))\n")
           val n3 = (n2+1 - n1) div 2 

           fun PHv(i,j) = Vector.tabulate 
                          (s, fn k => if i+1<n andalso
                                         NTset.sub(PM.sub(M,(i+1,j)),(0,k)) = 1
                                      then Vector.sub(PH,k) else NTrel.zero)
           val H = LG.Mat.tabulate 
                   (fn (i,j) => Vector.foldr NTrel.sum NTrel.zero (PHv(i,j)))

           fun PVv(i,j) = Vector.tabulate
                          (s, fn k => if 0 <= i-1 andalso
                                         NTset.sub(PM.sub(M,(j,i-1)),(0,k)) = 1
                                      then Vector.sub(PV,k) else NTrel.zero)
           val V = LG.Mat.tabulate 
                   (fn (i,j) => Vector.foldr NTrel.sum NTrel.zero (PVv(i,j)))

           val q = (n1+n3-1,n1+n3-1)
       in (H,V,q) end
   end
   local
       fun collect (1::bs) j acc = collect bs (j+1) (j::acc)
         | collect (_::bs) j acc = collect bs (j+1) acc
         | collect nil j acc = List.rev acc
   in
   fun NTsetToSyncats (X:NTset.t) = 
       let 
           val bits = List.tabulate (s, fn j => NTset.sub(X,(0,j)))
       in
           map G.nth (collect bits 0 [])
       end
   fun NTrelToSyncats (X:NTrel.t) =  
       let
           fun bits i = List.tabulate (s, fn j => NTrel.sub(X,(i,j)))
           fun row i = map (fn j => (G.nth i, G.nth j)) (collect (bits i) 0 [])
           val rows = List.tabulate (s,fn k => k)
       in 
           List.foldr (op @) [] (List.map row rows)
       end

   fun showPH () = Vector.map (fn r => NTrelToSyncats r) PH
   fun showPV () = Vector.map (fn r => NTrelToSyncats r) PV

   fun showPM(C:PM.t) : G.syncat list vector vector =
       let
           val n = Vector.length C
           val vtab = Vector.tabulate 
       in 
           vtab(n, fn i => vtab(n, fn j => NTsetToSyncats 
                                               (PM.sub(C,(i,j)))))
       end
   end

   fun showLG(H,V,q:int*int) =
       let
           val (n,_) = LG.Mat.dim
           fun tab f = Vector.tabulate (n,f)
           val sH = tab(fn i => tab(fn j => NTrelToSyncats 
                                                (LG.Mat.sub(H,(i,j)))))
           val sV = tab(fn i => tab(fn j => NTrelToSyncats 
                                                (LG.Mat.sub(V,(i,j)))))
       in 
           (sH,sV,q)
       end

   fun showDist Dist =
       let
           val (n,_) = LG.Mat.dim
           fun tab f = Vector.tabulate (n,f)
       in
           tab(fn i => tab(fn j => NTrelToSyncats 
                                       (LG.Mat.sub(Dist,(i,j)))))
       end
end
